/*
Copyright (c) 1998 The University of Texas
All Rights Reserved.
 
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
 
The GNU Public License is available in the file LICENSE, or you
can write to the Free Software Foundation, Inc., 59 Temple Place -
Suite 330, Boston, MA 02111-1307, USA, or you can find it on the
World Wide Web at http://www.fsf.org.
 
Programmers:	Niranjan Damera-Venkata, Thomas D. Kite, Brian L. Evans
		Extended by Wade Schwartzkopf to include PNG format
Version:        @(#)image_io.c	1.10	10/11/00

The authors are with the Laboratory for Image and Video Engineering
at The University of Texas at Austin, and can be reached at
{damera,tom,bevans}@vision.ece.utexas.edu.
*/

/* Standard includes */

#include <stdio.h>
#include <stdlib.h>

#include "image_io.h"
#include "readWritePPM.h"
#include "readWritePNG.h"

/* Allocate a raw byte image */
unsigned char* allocateImage(int numRows, int numColumns, int pixelSize)
{
    return((unsigned char*) calloc(numRows,
                                   numColumns*sizeof(unsigned char)*pixelSize));
}

/* Read 8 bit per pixel raw image */
static void readRawByteImage(char* filename, unsigned char *imgBufferPtr,
                             int *numRowsPtr, int *numColumnsPtr)
{
    FILE *fp = fopen(filename,"r");
    if (fp != NULL) {
        if (fread(imgBufferPtr, *numRowsPtr, *numColumnsPtr, fp) == 0) {
            fprintf(stderr, "Error reading file '%s'.\n", filename);
            exit(1);
        }
    }
    else {
        fprintf(stderr, "Error opening file '%s' for reading.\n", filename);
        exit(1);
    }
    fclose(fp);
}

/* Write 8 bit per pixel raw image data */
static void writeRawByteImage(char* filename, unsigned char *imgBufferPtr,
                              int *numRowsPtr, int *numColumnsPtr)
{
    FILE *fp = fopen(filename, "w");
    if (fp != NULL) {
        if ( fwrite(imgBufferPtr, *numRowsPtr,
                    (*numColumnsPtr) * sizeof(unsigned char), fp) == 0 ) {
            fprintf(stderr, "Error writing file '%s'.\n", filename);
            exit(1);
        }
    }
    else {
        fprintf(stderr, "Error opening file '%s' for writing.\n", filename);
        exit(1);
    }
    fclose(fp);
}

/*
Read header information from PGM or PNG formats and return
number of rows and columns.
*/
int readHeader(char* filename, int *numRowsPtr, int *numColumnsPtr)
{
   if (FileMatchPNG(filename, numRowsPtr, numColumnsPtr))
      return PNG;
   else
      return (FileMatchPPM(filename, numColumnsPtr, numRowsPtr));
}

/*
Read a byte image.  Could be in raw, PGM, or PNG formats.
Return image information cast as a char*.  Return
a null pointer on an error.
*/
int readByteImage(char* filename, unsigned char **imgBufferPtrPtr,
                  int *numRowsPtr, int *numColumnsPtr)
{
    int ppmType;

    /* Check to see if the filename is a PNG, PPM, or other type of file */
    if (!(ppmType = PNG*(!(!FileMatchPNG(filename, numRowsPtr, numColumnsPtr))))) {
       ppmType = FileMatchPPM(filename, numColumnsPtr, numRowsPtr); }
    switch (ppmType) {
      case RAW:
        if((*numRowsPtr>0) && (*numColumnsPtr>0)) {
           *imgBufferPtrPtr = allocateImage(*numRowsPtr, *numColumnsPtr, 1);
           readRawByteImage(filename, *imgBufferPtrPtr, numRowsPtr, numColumnsPtr);
        }
        else {
           fprintf(stderr,
                   "File '%s' is not PGM or PNG format.  "
                   "Reading raw image data requires valid arguments "
                   "for image height and width.\n", filename);
        }
        break;

      case PGM:
        *imgBufferPtrPtr = allocateImage(*numRowsPtr, *numColumnsPtr, 1);
        FileReadPPM(filename, *imgBufferPtrPtr, *numColumnsPtr, *numRowsPtr);
        break;

      case PPM:
        fprintf(stderr,
                "File '%s' is a color image, but color images "
                "are currently not supported.\n", filename);

      case PNG:
        *imgBufferPtrPtr = allocateImage(*numRowsPtr, *numColumnsPtr, 1);
        FileReadPNG(filename, *imgBufferPtrPtr);
        break;

      default:
        fprintf(stderr, "Unrecognized image type %d.\n", ppmType);
        break;
    }

    return(ppmType);
}

/* Write a byte image.  Could be in raw or PGM formats. */
void writeByteImage(char* filename, unsigned char *imgBufferPtr,
                    int *numRowsPtr, int *numColumnsPtr, int imageType)
{
    Tk_PhotoImageBlock block;

    switch (imageType) {
      case RAW:
        writeRawByteImage(filename, imgBufferPtr, numRowsPtr, numColumnsPtr);
        break;

      case PGM:
        InitImageInfo(&block, imgBufferPtr, imageType,
                      *numColumnsPtr, *numRowsPtr);
        FileWritePPM(filename, &block);
        break;

      case PPM:
        fprintf(stderr,
                "Request to write '%s' as a color image was not fulfilled "
                "because color images are currently not supported.\n",
                filename);

      case PNG:
	FileWritePNG(filename, imgBufferPtr, *numColumnsPtr, *numRowsPtr);
        break;

      default:
        fprintf(stderr, "Unrecognized image type %d.\n", imageType);
        break;
    }
}
