/*
Copyright (c) 2001 The University of Texas
All Rights Reserved.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

The GNU Public License is available in the file LICENSE, or you
can write to the Free Software Foundation, Inc., 59 Temple Place -
Suite 330, Boston, MA 02111-1307, USA, or you can find it on the
World Wide Web at http://www.fsf.org.

Programmers:    Wade Schwartzkopf
Version:        @(#)segmentation.c    1.0     02/24/00

The authors are with the Laboratory for Image and Video Engineering
at The University of Texas at Austin, and can be reached at
wade@.ece.utexas.edu.
*/

#include <math.h>
#include <stdlib.h>
#include "segmentation.h"
#include "my_matrix_stuff.h"
#include "imgtools.h"
#include "miscfunc.h"

int validCutLine(struct point start, struct point end, unsigned char** classImage, unsigned char** single, unsigned char** holes, int rows, int cols, int lineValue, double* lineArea, double lineHist[])
{
   int i,dx,dy,length;
   int xIndex, yIndex;
   float x,y,xinc,yinc;

   for(i = 0; i < 256; i++) {
      lineHist[i]=0;
   }

   /* Is cut line contained within object? */
   dx = end.x - start.x;
   dy = end.y - start.y;
   length = maxInt(abs(dx), abs(dy));
   *lineArea=0;
   xinc = ((float) dx)/((float) length);
   yinc = ((float) dy)/((float) length);

   x = start.x;
   y = start.y;
   for(i = 0; i <= length; i++) {
      yIndex=(int) rint(y);
      xIndex=(int) rint(x);
      if((single[yIndex][xIndex])<1) {
         if(!holes[yIndex][xIndex]) {
            return 0;
         }
      }
      else {
         (*lineArea)++;
         lineHist[classImage[yIndex][xIndex]]++;
      }
      single[yIndex][xIndex]=lineValue;  // Remove dividing line from object
      x+=xinc;
      y+=yinc;
   }

   return 1;
}

double calculateEntropy(unsigned char** inputImage, unsigned char** mask, int lowX, int highX, int lowY, int highY, int split, double lineArea, double lineHist[], double threshEntropy)
{
   int i,j;
   double area1=0, area2=lineArea;
   double hist1[256];
   double hist2[256];
   double entropy1=0.0, entropy2=0.0;

   /* Calculate histograms of the two objects */
   for(i = 0; i < 256; i++) {
      hist1[i]=0;
#ifdef INCLUDE_LINE
      if(lineArea>0) {
#else
      if(0) {
#endif
         hist2[i]=lineHist[i];
      }
      else {
         hist2[i]=0;
      }
   }
#ifndef INCLUDE_LINE
   area2=0;
   lineArea=0;
#endif
   for(i = lowY; i <= highY; i++) {
      for(j = lowX; j <= highX; j++) {
         if (mask[i][j]==1) {
            hist1[inputImage[i][j]]++;
            area1++;
         }
         else if (mask[i][j]>1) {
            hist2[inputImage[i][j]]++;
            area2++;
         }
      }
   }


   /* Calculate entropy */
   if(((area1<THRESH_SIZE)&&split)||
      (((area2-lineArea)<THRESH_SIZE)&&split)) {
      entropy1=(entropy2=MAXFLOAT);
   }
   else {
      for(i = 0; i < 256; i++) {
         if(hist1[i]>0) {
#ifndef PP_ENTROPY
            entropy1+=-(hist1[i]/area1)*log(hist1[i]/area1); 
#else
            entropy1+=-(hist1[i]/area1)*(hist1[i]/area1);
#endif

         }
         if((hist2[i]>0)&&split) {
#ifndef PP_ENTROPY
            entropy2+=-(hist2[i]/area2)*log(hist2[i]/area2);
#else
            entropy2+=-(hist2[i]/area2)*(hist2[i]/area2);
#endif
         }
      }
   }
#ifdef PP_ENTROPY
   entropy1+=1;
   if(split) {
      entropy2+=1;
   }
#endif
//   printf("Entropy: %f\r\n", entropy1+entropy2);

   if(split) {
      if(maxDouble(entropy1, entropy2)>(threshEntropy-THRESH_E_STEP)) { return MAXFLOAT; }
   }

#ifndef PP_ENTROPY
   return entropy1 + entropy2;
#else
   return (entropy1 + entropy2)/(split+1);
#endif
}

struct pointList* selectCutLine(struct pointList* boundary_points, unsigned char** inputImage, unsigned char** single, unsigned char** labels, unsigned char** holes, int* nextClass, int rows, int cols)
{
   struct pointList* tempptr_i;
   struct pointList* tempptr_j;
   struct point best1 = {0, 0};
   struct point best2 = {0, 0};
   unsigned char** tempSingle=allocateByteMatrix(rows, cols);
   unsigned char** blank=allocateByteMatrix(rows, cols);
   double tempEntropy;
   double initialEntropy;
   double bestEntropy=MAXFLOAT;
   double lineArea;
   double lineHist[256];
   int lowX=cols, highX=0, lowY=rows, highY=0;
   int i,j;

   clearImage(blank, 0, cols-1, 0, rows-1);
   copy_image(single, tempSingle, rows, cols);
   // Calculate bounding box around object to avoid unecessary calculations
   //    outside of this box.
   for(i = 0; i < rows; i++) {
      for(j = 0; j < cols; j++) {
         if((i<lowY)&&(single[i][j]>0)) { lowY=i; }
         if((i>highY)&&(single[i][j]>0)) { highY=i; }
         if((j<lowX)&&(single[i][j]>0)) { lowX=j; }
         if((j>highX)&&(single[i][j]>0)) { highX=j; }
      }
   }

   initialEntropy=calculateEntropy(inputImage, tempSingle, lowX, highX, lowY, highY, 0, 0, NULL, 0);
   bestEntropy=initialEntropy;
   printf("Initial entropy: %f | ", initialEntropy);
   if(initialEntropy==0.0) { 
      printf("Final entropy: %f | ", initialEntropy);
      freeByteMatrix(tempSingle);
      freeByteMatrix(blank);
      return NULL;
   }
 
   // Examine every possible combination of cut points
   i=1;
   for(tempptr_i = boundary_points; tempptr_i->next; tempptr_i = tempptr_i->next) {
      for(tempptr_j = tempptr_i->next; tempptr_j; tempptr_j = tempptr_j->next) {
         copy_square(single, tempSingle, lowX, highX, lowY, highY);
         if(validCutLine(tempptr_i->value, tempptr_j->value, inputImage, tempSingle, holes, rows, cols, 0, &lineArea, lineHist)) {
            if(labelClusters(tempSingle, blank, lowX, highX, lowY, highY, 4)==2) {
               if((tempEntropy=calculateEntropy(inputImage, blank, lowX, highX, lowY, highY, 1, lineArea, lineHist, initialEntropy)) < bestEntropy) {
                  bestEntropy = tempEntropy;
                  best1 = tempptr_i->value;
                  best2 = tempptr_j->value;
               }
            }
         } 
      }
      //printf("Point #%d: x=%d, y=%d; Minimum entropy: %f at ([%d %d], [%d %d])\r\n", i++, tempptr_i->value.x, tempptr_i->value.y, bestEntropy, best1.x+1, best2.x+1, best1.y+1, best2.y+1);
   }
   printf("Final average entropy: %f | ", bestEntropy);

   if(bestEntropy>=((initialEntropy)-THRESH_E_STEP)) { 
      // Do nothing
      freeByteMatrix(tempSingle);
      freeByteMatrix(blank);
      return NULL;
   }
   else {
      // Split object into two objects
      copy_square(single, tempSingle, lowX, highX, lowY, highY);
      validCutLine(best1, best2, inputImage, tempSingle, holes, rows, cols, 0, &lineArea, lineHist);
      labelClusters(tempSingle, blank, lowX, highX, lowY, highY, 4);
      relabel(blank, labels, 2, *nextClass, lowX, highX, lowY, highY);
      validCutLine(best1, best2, inputImage, labels, holes, rows, cols, 0, &lineArea, lineHist);
      (*nextClass)++;
      freeByteMatrix(tempSingle);
      freeByteMatrix(blank);
      return addPoint(addPoint(NULL, best1.x, best1.y), best2.x, best2.y);
   }
}

int isCluster(unsigned char** inputImage, unsigned char** labels, unsigned char** single, int objectLabel, int lowX, int highX, int lowY, int highY)
{
   find_equal(labels, single, objectLabel, lowX, highX, lowY, highY);
   return (calculateEntropy(inputImage, single, lowX, highX, lowY, highY, 0, 0, NULL, 0)>THRESH_ENTROPY);
}

void filterBoundaryPoints(unsigned char** inputImage, struct pointList* boundary, int rows, int cols)
{
   int i,j,k;
   int remove;
   const int connectedX[] = X_NEIGHBORS;
   const int connectedY[] = Y_NEIGHBORS;
   const int CONNECTEDNESS = 8;
   struct pointList *tempptr_i, *prevptr;

   // Only use boundary points that are next to a different class
   prevptr=boundary;
   if (boundary) {
      for(tempptr_i = boundary->next; tempptr_i; tempptr_i = tempptr_i->next) {
         remove=1;
         i=tempptr_i->value.y;
         j=tempptr_i->value.x;
         for(k = 0; k < CONNECTEDNESS; k++) {
            if (((j+connectedX[k])>=0)&&((j+connectedX[k])<cols)&&
                ((i+connectedY[k])>=0)&&((i+connectedY[k])<rows)&&
                ((inputImage[i+connectedY[k]][j+connectedX[k]])>0)&&
                ((inputImage[i+connectedY[k]][j+connectedX[k]])!=(inputImage[i][j]))) {
               remove=0;
            }
         }
         if(remove) {
            prevptr->next=tempptr_i->next;
            free(tempptr_i);
            tempptr_i=prevptr;
         }
         prevptr=tempptr_i;
      }
   }
}

void equivalent(unsigned char** inputImage, unsigned char** mask, struct labelList* labelsToMerge, int lowX, int highX, int lowY, int highY)
{
   int merged=1;
   int label1=0, label2=0;
   struct labelList* tempCopy;
   struct labelList* tempptr_i;
   struct labelList* tempptr_j;
   double entropy1, entropy2, jointEntropy;
   double newRatio, bestRatio=1.0;
   unsigned char** temp=allocateByteMatrix(highY+1, highX+1);

   tempCopy=copyLabels(labelsToMerge);
   while(merged) {
      merged=0;
      bestRatio=1.0;
      for(tempptr_i = tempCopy; tempptr_i->next; tempptr_i = tempptr_i->next) {
         for(tempptr_j = tempptr_i->next; tempptr_j; tempptr_j = tempptr_j->next) {
            find_equal(mask, temp, tempptr_i->value, lowX, highX, lowY, highY);
            entropy1=calculateEntropy(inputImage, temp, lowX, highX, lowY, highY, 0, 0, NULL, 0);
            printf("Object %d entropy = %1.2f | ", tempptr_i->value, entropy1);
            find_equal(mask, temp, tempptr_j->value, lowX, highX, lowY, highY);
            entropy2=calculateEntropy(inputImage, temp, lowX, highX, lowY, highY, 0, 0, NULL, 0);
            printf("Object %d entropy = %1.2f | ", tempptr_j->value, entropy2);
            find_two_equal(mask, temp, tempptr_i->value, tempptr_j->value, lowX, highX, lowY, highY);
            jointEntropy=calculateEntropy(inputImage, temp, lowX, highX, lowY, highY, 0, 0, NULL, 0);
            printf("Joint entropy = %1.4f\r\n", jointEntropy);
            if((jointEntropy<=maxDouble(entropy1, entropy2))||(jointEntropy<0.01)) {
               merged=1;
               newRatio=jointEntropy;
               if(newRatio<bestRatio) {
                  bestRatio=newRatio;
                  label1=tempptr_i->value;
                  label2=tempptr_j->value;
               }
            }
         }
      }
      if(merged) {
         printf("Joining objects %d and %d...\r\n", label1, label2);
         relabel(mask, mask, label2, label1, lowX, highX, lowY, highY);
         skiplabel(mask, label2, lowX, highX, lowY, highY);
         tempCopy=removeLabel(label2, tempCopy);
      }
   }

   freeLabelList(tempCopy);
   freeByteMatrix(temp);
}


void segment(unsigned char** inputImage, unsigned char** outputImage, int rows, int cols)
{
   //FILE* cut_lines;
   int nextClass, clusters;
   int i=0;
   unsigned char** initialLabels=allocateByteMatrix(rows, cols);
   unsigned char** labels=allocateByteMatrix(rows, cols);
   unsigned char** temp=allocateByteMatrix(rows, cols);
   unsigned char** holes=allocateByteMatrix(rows, cols);
   unsigned char** boundary=allocateByteMatrix(rows, cols);
   unsigned char** single=allocateByteMatrix(rows, cols);
   unsigned char** clusterParts=allocateByteMatrix(rows, cols);
   struct pointList* boundaryPoints=NULL;
   struct pointList* cutPoints=NULL;
   struct labelList* clusterLabels=NULL;

   // Label connected components
   clusters = labelClusters(inputImage, labels, 0, cols-1, 0, rows-1, 4);
   nextClass = clusters + 1;
   copy_image(labels, initialLabels, rows, cols);
   //writeImageAs2DByte("images/labels", labels, rows, cols, 0);

   // Locate holes in clusters
   findHoles(inputImage, holes, 0, cols-1, 0, rows-1);
   //writeImageAs2DByte("images/holes", holes, rows, cols, 0);

   // Separate touching chromosomes
   //cut_lines=fopen("images/cutlines", "w");
   for(i = 1; ((i < nextClass)&&(i<256)); i++) {
      cutPoints=(struct pointList*) 1;
      printf("Checking chromosome %d of %d...\r\n", i, nextClass-1);
      while(isCluster(inputImage, labels, single, i, 0, cols-1, 0, rows-1)&&cutPoints) {
         findHoles(labels, holes, 0, cols-1, 0, rows-1);
         findBoundary(single, boundary, holes, rows, cols);
         boundaryPoints=listPoints(boundary, rows, cols);
         filterBoundaryPoints(inputImage, boundaryPoints, rows, cols);
         cutPoints=selectCutLine(boundaryPoints, inputImage, single, labels, holes, &nextClass, rows, cols);
         //writeImageAs2DByte("images/labels2", labels, rows, cols, 0);
         printf("Object #%d\r\n", i);
         printPointList(cutPoints);
         if(cutPoints) {
            //fprintf(cut_lines, "line([%d %d], [%d %d]);\r\n", cutPoints->value.x+1, cutPoints->next->value.x+1, cutPoints->value.y+1, cutPoints->next->value.y+1);
            freePointList(cutPoints);
            cutPoints=(struct pointList*) 1;
         }
      }
   }
   //fclose(cut_lines);

   // Associate ends of overlapping chromosomes
   for(i = 1; i < clusters; i++) {
      printf("Checking cluster %d of %d...\r\n", i, clusters-1);
      find_equal(initialLabels, single, i, 0, cols-1, 0, rows-1);
      ifthen(single, labels, clusterParts, rows, cols);
      clusterLabels=listLabels(clusterParts, rows, cols);
      equivalent(inputImage, labels, clusterLabels, 0, cols-1, 0, rows-1);
      if(clusterLabels) {
         freeLabelList(clusterLabels);
      }
   }

   copy_image(labels, outputImage, rows, cols);

   freeByteMatrix(initialLabels);
   freeByteMatrix(labels);
   freeByteMatrix(temp);
   freeByteMatrix(holes);
   freeByteMatrix(boundary);
   freeByteMatrix(single);
   freeByteMatrix(clusterParts);
   freePointList(boundaryPoints);
}
